/* 
   This file dumps a trace from a Lecroy 9450 oscilloskope via GPIB 
   gcc get_trace_9450.c -o get_trace_9450 -lgpib -Wall      
   
   changelog:
   12.09.2009 remove warnings
   05.10.2007 change from NI-GPIB to "linux gpib" driver. 
   03.10.2006 adapt to LeCroy 9450 Waverunner
   

   FIXME: 
   - inconsistency in byte ordering?
   - remove unused tmp files.  
   
   Niels Bassler <bassler@phys.au.dk>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gpib/ib.h>

#define BUFFER_SIZE 0xffff
#define USAGE_STRING "Usage: %s GPIBaddress [channel [outputfile]]\n" 
#define TEMP_FILE_NAME "/tmp/TMPLeCroyData_XXXXXX"

#define DEBUG 0
#define DESCRIPTOR_SIZE 346   /* length of descriptor block in bytes */
#define RT_SINGLE_SWEEP 0     /* single sweep mode identifier */
#define COMM_ORDER_LO   0     /* MSB before LSB , LSB before MSB ... fixme*/
#define COMM_TYPE_WORD  1     /* 16 bit  data */
#define COMM_TYPE_BYTE  0     /* 8 bit   data */

#define FORMAT_COMMAND  "COMM_FORMAT DEF9, BYTE, BIN"
//#define FORMAT_COMMAND  "COMM_FORMAT DEF9, WORD, BIN" // for 16 bit data

typedef struct _time_stamp{
  double seconds;
  char   minutes;
  char   hours;
  char   days;
  char   months;
  short  year;
  short  unused;
} ttime_stamp;

typedef struct _ResponseChunk{
  char   QueryText[10];
  char   Query2[2];
  char   QyeryLength[9];
} __attribute((packed)) tResponseChunk;

typedef struct _DescriptorChunk {
  /* words are always transferred with the most significant byte first */
  char       DescriptorName[16];
  char       TemplateName[16];
  //char       NotUsed_0;
  short      CommType;
  //char       NotUsed_1;
  short      CommOrder;
  
  long       WaveDescriptor;
  long       UserText;
  long       ResDesc1;
  
  long       TrigtimeArray; 
  long       RisTimeArray;
  long       ResArray1;
  long       WaveArray1;
  long       WaveArray2;
  long       ResArray2;
  long       ResArray3;
  
  char       InstrumentName[16];
  long       InstrumentNumber;
  
  char       TraceLabel[16];
  short      Reserved1;
  short      Reserved2;
  
  long       WaveArrayCount;
  long       PntsPerScreen;
  long       FirstValidPnt;
  long       LastValidPnt;
  long       FirstPoint;
  long       SparsingFactor;
  long       SegmentIndex;
  long       SubarrayCount;
  long       SweepsPerAcq;
  long       Obsolete1;
  float      VerticalGain;
  float      VerticalOffset;
  
  float      MaxValue;
  float      MinValue;

  short      NominalBits;
  short      NomSubarrayCount;
  float      HorizInterval;
  double     HorizOffset;
  double     PixelOffset;
  
  char       Vertunit[48];
  char       Horunit[48];
  short      Reserved3;
  short      Reserved4;
  
  ttime_stamp  TriggerTime;
  
  float      AcqDuration;
  //  char       NotUsed_3;
  short      RecordType;
  //  char       NotUsed_4;
  short      ProcessingDone;
  short      Reserved5;
  short      RisSweeps;

  //char       NotUsed_5;
  short      Timebase;

  //  char       NotUsed_6;
  short      VertCoupling;

  float      ProbeAtt;

  //char       NotUsed_7;
  short      FixedVertGain;
  
  //char       NotUsed_8;
  short      BandwidthLimit;
  
  float      VerticalVernier;
  float      AcqVertOffset;

  //  u_int8_t   NotUsed_9;
  
  short       WaveSource;
} __attribute((packed)) tDescriptorChunk;

/* usertext block follows */
/* data array 1 follows */


/* fixed vertical gain in V/div */
const float FixedVertGain[] = { 
  1e-6, 2e-6, 5e-6,
  1e-5, 2e-5, 5e-5,
  1e-4, 2e-4, 5e-4,
  1e-3, 2e-3, 5e-3,
  1e-2, 2e-2, 5e-2,
  1e-1, 2e-1, 5e-1,
  1, 2, 5,
  1e1, 2e1, 5e1,
  1e2, 2e2, 5e2,
  1e3
};

const float time_base[] = {
  1e-12, 2e-12, 5e-12,
  1e-11, 2e-11, 5e-11,
  1e-10, 2e-10, 5e-10,
  1e-09, 2e-09, 5e-09,
  1e-08, 2e-08, 5e-08,
  1e-07, 2e-07, 5e-07,
  1e-06, 2e-06, 5e-06,
  1e-05, 2e-05, 5e-05,
  1e-04, 2e-04, 5e-04,
  1e-03, 2e-03, 5e-03,
  1e-02, 2e-02, 5e-02,
  1e-01, 2e-01, 5e-01,
  1e00, 2e00, 5e00,
  1e01, 2e01, 5e01,
  1e02, 2e02, 5e02,
  1e03, 2e03, 5e03
};


void gpiberr(char *msg);
void swap_16(char unsigned *x);
float convert_SampleInterval( char val);

int main(int argc, char *argv[]) {

  //  double vertical_offset, variable_gain, fixed_gain, fdata;
  double fdata;

  //  long datachunk_size = 0;
  long total_size = 0;
  int address, channel = 1, i, fd;
  char command[BUFFER_SIZE], buffer[BUFFER_SIZE];
  char tmp_filename[] = TEMP_FILE_NAME;
  FILE *fp;

  tDescriptorChunk Descriptor;
  //  ttime_stamp time_stamp;
  tResponseChunk Response;

  char unknown[22]; /* unknown stuff */

  //  char Header[4];
  //  char *Data;
  char *Usertext;
  short *Wavearray;  /* 16 bit data */
  char *sWavearray;  /* 8 bit data  */

  if ((argc > 4) || (argc < 2)){
    printf(USAGE_STRING, argv[0]);
    exit(0);
  } 

  address =  atoi(argv[1]);
  /* if no channel selected, use default C1 */
  if (argc >= 3) {
    if (atoi(argv[2]) > 4) {
      fprintf(stderr,"Invalid channel number\n");
      exit(-1);
    }
    channel = atoi(argv[2]);
    if ((channel > 4) || (channel < 1)) {
      printf("*** ERROR: No such channel.\n");
      exit(-1);
    }
  }

  if(DEBUG) {
    printf("Descriptor size: %i\n",sizeof(Descriptor));
    
    //intf(" size: %i\n",sizeof(Descriptor.WaveDescriptor));
  }

  if ( sizeof(Descriptor) != DESCRIPTOR_SIZE) {
    printf(" *** ERROR: allocated descriptor has wrong size. This is a platform problem, and needs some reprogramming to be fixed.\n");
    exit(-1);
  }


  for (i = 0; i < BUFFER_SIZE; i++)
    buffer[i] = 0;

  SendIFC(0);
  if (ibsta & ERR ) {
    gpiberr(" ERR");
    exit(0);
  }

  /* set format */
  strcpy(command, FORMAT_COMMAND);
  Send(0,
       (Addr4882_t) address,
       command,
       (long) strlen(command),
       NLend);

  /* LSB first */
  strcpy(command, "COMM_ORDER LO");
  Send(0,
       (Addr4882_t) address,
       command,
       (long) strlen(command),
       NLend);
/*   Receive(0, (Addr4882_t) address, buffer, */
/* 	  (long) BUFFER_SIZE, STOPend); */

  
  if (channel == 3)
    strcpy(command, "MC:WAVEFORM?");
  if (channel == 4)
    strcpy(command, "MD:WAVEFORM?");
  if (channel < 3)
    sprintf(command, "C%i:WAVEFORM?", channel);
  

  /* generate a tempfilename */
  fd = mkstemp(tmp_filename);
  if (fd == -1) {
    fprintf(stderr, "Cannot create tempfile\n");
    exit(1);
  }
  close(fd);

  Send(0, 
       (Addr4882_t) address, 
       command, 
       (long) strlen(command), 
       NLend);

  /* open a target file */
  fp = fopen(tmp_filename, "w");
  if (fp==NULL) {
    fprintf(stderr, "ERROR: %s cant create output file.",argv[0]);
    ibonl (0,0);
    exit(-1);
  }

  do { /* get all data until END is set */
    Receive(0, (Addr4882_t) address, buffer, 
	    (long) BUFFER_SIZE, STOPend);
    if(DEBUG)
      printf(" ibcnt : %i - ibsta: %0xi\n", ibcnt, ibsta);    
    total_size += ibcnt;
    for (i = 0; i < ibcnt; i++)
      putc(buffer[i], fp);
  }
  while (!(ibsta & END));

  fclose(fp);
  if (DEBUG) 
    printf("Wrote %li bytes to %s.\n", total_size, tmp_filename);

  /* ************************************** */
  //  exit(0);

  /* reopen all the stuff with read access, and get the descriptor and data */
  fp = fopen(tmp_filename, "r");
  fread( &Response, sizeof(Response), (size_t) 1, fp);
  fread( &Descriptor, sizeof(Descriptor), (size_t) 1, fp);

  if (DEBUG) {
    printf(" DN: %s\n", Descriptor.DescriptorName);
    printf(" TN: %s\n", Descriptor.TemplateName);
    printf(" CT: %i\n", Descriptor.CommType);
    printf(" CO: %i\n", Descriptor.CommOrder);
    printf(" WD: %li\n", Descriptor.WaveDescriptor);
    printf(" UT: %li\n", Descriptor.UserText);
    printf(" RD1: %li\n", Descriptor.ResDesc1);
    printf(" TA: %li\n", Descriptor.TrigtimeArray);
    printf(" RTA: %li\n", Descriptor.RisTimeArray);
    printf(" RA1: %li\n", Descriptor.ResArray1);
    printf(" WA1: %li\n", Descriptor.WaveArray1);
    printf(" WA2: %li\n", Descriptor.WaveArray2);
    printf(" RA2: %li\n", Descriptor.ResArray2);
    printf(" RA3: %li\n", Descriptor.ResArray3);
    printf(" IN: %s\n", Descriptor.InstrumentName);
    printf(" I#: %li\n", Descriptor.InstrumentNumber);
    printf(" TL: %s\n\n", Descriptor.TraceLabel);
    
    printf(" WAC: %li\n", Descriptor.WaveArrayCount);
    printf(" PPS: %li\n", Descriptor.PntsPerScreen);
    printf(" FVP: %li\n", Descriptor.FirstValidPnt);
    printf(" LVP: %li\n\n", Descriptor.LastValidPnt);
    
    printf(" VG: %e\n", Descriptor.VerticalGain);
    printf(" VO: %e\n", Descriptor.VerticalOffset);
    printf(" VU: %s\n\n", Descriptor.Vertunit);

    printf(" VG: %e\n", Descriptor.HorizInterval);
    printf(" VO: %e\n", Descriptor.HorizOffset);
    printf(" HU: %s\n\n", Descriptor.Horunit);

    printf(" RT: %i\n\n", Descriptor.RecordType);

  }
  fread( &unknown, sizeof(unknown), (size_t) 1, fp);  
  
  /* allocate memory for Usertext */
  Usertext = (char *) calloc(Descriptor.UserText, sizeof(char));
  fread( Usertext, Descriptor.UserText, (size_t) 1, fp);
  free(Usertext); /* but we dont need it */
  
  /* allocate memory for Wavearray */

  Wavearray = NULL;
  sWavearray = NULL;

  if (Descriptor.CommType == COMM_TYPE_WORD) {
    Wavearray = calloc(Descriptor.WaveArray1, sizeof(char));
    fread( Wavearray, Descriptor.WaveArray1, (size_t) 1, fp);
  }
  else if (Descriptor.CommType == COMM_TYPE_BYTE) {
    sWavearray = calloc(Descriptor.WaveArray1, sizeof(char));
    fread( sWavearray, Descriptor.WaveArray1, (size_t) 1, fp);
  }
  fclose(fp);
  unlink(tmp_filename);
  //  exit(0);

  /* now lets parse the relevant header information  */
  /* write output data */
  if (argc == 4) /* print to file*/
    fp = fopen(argv[3],"w");
  else
    fp = fopen("out.dat", "w");
  
  if (Descriptor.CommOrder != COMM_ORDER_LO ) {
    fprintf(stderr, "ERR: Osc. comm order wrong (LSB should be first)\n");
    //exit(-1);
  }
  if (Descriptor.RecordType != RT_SINGLE_SWEEP) {
    fprintf(stderr, "ERR: not single sweep mode\n");
    exit(-1);
  }

  if (Descriptor.LastValidPnt > Descriptor.WaveArray1) {
    fprintf(stderr, "ERR: last point beyond array scope.");
    exit(-1);
  }
  for(i=Descriptor.FirstValidPnt; i <= Descriptor.LastValidPnt; i++) {
    if(Descriptor.CommType == COMM_TYPE_WORD)
      fdata = Descriptor.VerticalGain * Wavearray[i] - Descriptor.VerticalOffset;
    else 
       fdata = Descriptor.VerticalGain * sWavearray[i] - Descriptor.VerticalOffset;
    fprintf(fp, "%e %e\n", i * Descriptor.HorizInterval + Descriptor.HorizOffset, fdata);
  }  
  fclose(fp);
   if (Descriptor.CommType == COMM_TYPE_WORD)
     free(Wavearray);
   else 
     free(sWavearray);

  return(0);
}

void gpiberr(char *msg) {

    fprintf (stderr,"%s\n", msg);

    fprintf (stderr,"ibsta = &H%x  <", ibsta);
    if (ibsta & ERR )  fprintf (stderr," ERR");
    if (ibsta & TIMO)  fprintf (stderr," TIMO");
    if (ibsta & END )  fprintf (stderr," END");
    if (ibsta & SRQI)  fprintf (stderr," SRQI");
    if (ibsta & RQS )  fprintf (stderr," RQS");
    if (ibsta & SPOLL) fprintf (stderr," SPOLL");
    if (ibsta & EVENT) fprintf (stderr," EVENT");
    if (ibsta & CMPL)  fprintf (stderr," CMPL");
    if (ibsta & LOK )  fprintf (stderr," LOK");
    if (ibsta & REM )  fprintf (stderr," REM");
    if (ibsta & CIC )  fprintf (stderr," CIC");
    if (ibsta & ATN )  fprintf (stderr," ATN");
    if (ibsta & TACS)  fprintf (stderr," TACS");
    if (ibsta & LACS)  fprintf (stderr," LACS");
    if (ibsta & DTAS)  fprintf (stderr," DTAS");
    if (ibsta & DCAS)  fprintf (stderr," DCAS");
    fprintf (stderr," >\n");
 
    fprintf (stderr,"iberr = %d", iberr);
    if (iberr == EDVR) fprintf (stderr," EDVR <DOS Error>\n");
    if (iberr == ECIC) fprintf (stderr," ECIC <Not CIC>\n");
    if (iberr == ENOL) fprintf (stderr," ENOL <No Listener>\n");
    if (iberr == EADR) fprintf (stderr," EADR <Address error>\n");
    if (iberr == EARG) fprintf (stderr," EARG <Invalid argument>\n");
    if (iberr == ESAC) fprintf (stderr," ESAC <Not Sys Ctrlr>\n");
    if (iberr == EABO) fprintf (stderr," EABO <Op. aborted>\n");
    if (iberr == ENEB) fprintf (stderr," ENEB <No GPIB board>\n");
    if (iberr == EOIP) fprintf (stderr," EOIP <Async I/O in prg>\n");
    if (iberr == ECAP) fprintf (stderr," ECAP <No capability>\n");
    if (iberr == EFSO) fprintf (stderr," EFSO <File sys. error>\n");
    if (iberr == EBUS) fprintf (stderr," EBUS <Command error>\n");
    if (iberr == ESTB) fprintf (stderr," ESTB <Status byte lost>\n");
    if (iberr == ESRQ) fprintf (stderr," ESRQ <SRQ stuck on>\n");
    if (iberr == ETAB) fprintf (stderr," ETAB <Table Overflow>\n");
 
    fprintf (stderr,"ibcnt = %d\n", ibcnt);
    fprintf (stderr,"\n");
    ibonl (0,0);
    exit(1);

}

void swap_16 (char unsigned *x) {

  unsigned char c;
  
  c = *x;
  *x = *(x+1);
  *(x+1) = c;
  return;
}

